; Calculating Bezier surfaces;
; Implemented in Flat assembler by Maciej Guba,
; http://macgub.hekko.pl
; I use this equation:
; x(s,t)=(1-s)^3*(x11*a+x12*b+x13*c+x14*d)+
;	 +3*(1-s)^2*s*(x21*a+x22*b+x23*c+x24*d)+
;	 +3*(1-s)*s^2*(x31*a+x32*b+x33*c+x34*d)+
;	 +s^3*(x41*a+x42*b+x43*c+x44*d)
;
;  a = (1-t)^3
;  b = 3*(1-t)^2*t
;  c = 3*(1-t)*t^2
;  d = t^3




calc_bezier_surface:
; calc single surface, write vertices to [edi],
; faces to [ebx]
;  in  :
;	esi - ptr  -> x11, y11, z11 .... x44, y44, z44 ; dword
;	edx - quality -> s shl 16 t
;	edi - ptr to area of calculated vertices
;	ebx - ptr to area of calculated faces (triangles)
;  out :
;	ecx - vertices count
;	edx - triangles count

	push  ebp
	mov   ebp,esp
	sub   esp,512
	sub   ebp,16
	and   ebp,0xfffffff0

  .t equ [ebp-4]
  .s equ [ebp-8]
  .ct equ	  dword [ebp-12] ; integer counter t
  .cs equ	  dword [ebp-16] ; int counter s
  .vert_count equ dword [ebp-20]
  .tri_count  equ dword [ebp-24]

  .x4 equ ebp-48	;      .x11-x14
  .x3 equ ebp-64	;      .x21-x24
  .x2 equ ebp-80	;      .x31-x34
  .x1 equ ebp-96	;      .x41-x44
  .y4 equ ebp-112	;      .y11-y14
  .y3 equ ebp-128	;      .y21-y24
  .y2 equ ebp-144	;      .y31-y34
  .y1 equ ebp-160	;      .y41-y44
  .z4 equ ebp-176	;      .z11-z14
  .z3 equ ebp-192	;      .z21-z24
  .z2 equ ebp-208	;      .z31-z34
  .z1 equ ebp-224	;      .z41-z44
  .ds equ  [ebp-228]	; delta s
  .dt equ  [ebp-232]	; delta t
  .d  equ  [ebp-244]
  .c  equ  [ebp-248]
  .b  equ  [ebp-252]
  .a  equ  [ebp-256]
 ; .dtt equ [ebp-260]
 ; .ctt equ [ebp-264]
 ; .btt equ [ebp-268]
 ; .att equ [ebp-272]


   push dword[esi]   ; first convert derives for easier calculating.
   pop	dword[.x1]
   push dword[esi+12]
   pop	dword[.x1+4]
   push dword[esi+24]
   pop	dword[.x1+8]
   push dword[esi+36]
   pop	dword[.x1+12]

   push dword[esi+36+12]    ; shuffle x
   pop	dword[.x2]
   push dword[esi+36+12*2]
   pop	dword[.x2+4]
   push dword[esi+36+12*3]
   pop	dword[.x2+8]
   push dword[esi+36+12*4]
   pop	dword[.x2+12]

   push dword[esi+36+12*5]
   pop	dword[.x3]
   push dword[esi+36+12*6]
   pop	dword[.x3+4]
   push dword[esi+36+12*7]
   pop	dword[.x3+8]
   push dword[esi+36+12*8]
   pop	dword[.x3+12]

   push dword[esi+36+12*9]
   pop	dword[.x4]
   push dword[esi+36+12*10]
   pop	dword[.x4+4]
   push dword[esi+36+12*11]
   pop	dword[.x4+8]
   push dword[esi+36+12*12]
   pop	dword[.x4+12]

   push dword[esi+4]	      ;shuffle y
   pop	dword[.y1]
   push dword[esi+12+4]
   pop	dword[.y1+4]
   push dword[esi+24+4]
   pop	dword[.y1+8]
   push dword[esi+36+4]
   pop	dword[.y1+12]

   push dword[esi+36+12+4]
   pop	dword[.y2]
   push dword[esi+36+12*2+4]
   pop	dword[.y2+4]
   push dword[esi+36+12*3+4]
   pop	dword[.y2+8]
   push dword[esi+36+12*4+4]
   pop	dword[.y2+12]

   push dword[esi+36+12*5+4]
   pop	dword[.y3]
   push dword[esi+36+12*6+4]
   pop	dword[.y3+4]
   push dword[esi+36+12*7+4]
   pop	dword[.y3+8]
   push dword[esi+36+12*8+4]
   pop	dword[.y3+12]

   push dword[esi+36+12*9+4]
   pop	dword[.y4]
   push dword[esi+36+12*10+4]
   pop	dword[.y4+4]
   push dword[esi+36+12*11+4]
   pop	dword[.y4+8]
   push dword[esi+36+12*12+4]
   pop	dword[.y4+12]


   push dword[esi+8]	      ;shuffle z
   pop	dword[.z1]
   push dword[esi+12+8]
   pop	dword[.z1+4]
   push dword[esi+24+8]
   pop	dword[.z1+8]
   push dword[esi+36+8]
   pop	dword[.z1+12]

   push dword[esi+36+12+8]
   pop	dword[.z2]
   push dword[esi+36+12*2+8]
   pop	dword[.z2+4]
   push dword[esi+36+12*3+8]
   pop	dword[.z2+8]
   push dword[esi+36+12*4+8]
   pop	dword[.z2+12]

   push dword[esi+36+12*5+8]
   pop	dword[.z3]
   push dword[esi+36+12*6+8]
   pop	dword[.z3+4]
   push dword[esi+36+12*7+8]
   pop	dword[.z3+8]
   push dword[esi+36+12*8+8]
   pop	dword[.z3+12]

   push dword[esi+36+12*9+8]
   pop	dword[.z4]
   push dword[esi+36+12*10+8]
   pop	dword[.z4+4]
   push dword[esi+36+12*11+8]
   pop	dword[.z4+8]
   push dword[esi+36+12*12+8]
   pop	dword[.z4+12]


   movzx    ecx,dx
   mov	    .cs,ecx
 ;  sub      .cs,1
   cvtsi2ss xmm1,ecx
   rcpss    xmm1,xmm1
   movss    .ds,xmm1
   shr	    edx,16
   mov	    .ct,edx
;   sub      .ct,1
   cvtsi2ss xmm0,edx
   rcpss    xmm0,xmm0
   movss    .dt,xmm0

 ;  push     edi
   push     ebx

   xor	    ecx,ecx
 ;  mov      .s,ecx
   mov	    .t,ecx
   mov	    .vert_count,ecx
   mov	    .tri_count,ecx

 .looptt:
   push   ecx
   movss  xmm1,[f1]
   subss  xmm1,.t
   movss  xmm2,xmm1
   movss  xmm3,xmm1
   movss  xmm4,.t
   mulss  xmm2,xmm1
   mulss  xmm1,[f3]
   mulss  xmm3,xmm2
   movss  .a,xmm3
   mulss  xmm2,[f3]
   mulss  xmm4,xmm4
   mulss  xmm2,.t
   movss  .b,xmm2
   mulss  xmm1,xmm4
   movss  .c,xmm1
   mulss  xmm4,.t
   movss  .d,xmm4

   xor	    ecx,ecx
   mov	    dword .s,ecx

 .loopss:
   push  ecx
   movss xmm0,[f1]
   subss xmm0,.s
   movss xmm3,.s
   movss xmm1,xmm0
   mulss xmm1,xmm3
   mulss xmm1,[f3]
   movss xmm2,xmm1
   mulss xmm1,xmm0  ; xmm1 - bs
   mulss xmm2,xmm3  ; xmm2 - cs
   movss xmm4,xmm0
   mulss xmm0,xmm0
   mulss xmm0,xmm4  ; xmm0 - as
   mulss xmm3,xmm3
   mulss xmm3,.s    ; xmm3 - ds

   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.x1]
   mulps    xmm5,[.x2]
   mulps    xmm6,[.x3]
   mulps    xmm7,[.x4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi],xmm4


   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.y1]
   mulps    xmm5,[.y2]
   mulps    xmm6,[.y3]
   mulps    xmm7,[.y4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+4],xmm4

   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.z1]
   mulps    xmm5,[.z2]
   mulps    xmm6,[.z3]
   mulps    xmm7,[.z4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+8],xmm4


   movss    xmm0,.s
   addss    xmm0,.ds
   movss    .s,xmm0
   add	    edi,12
   add	    .vert_count,1
   pop	    ecx
   add	    ecx,1
   cmp	    ecx,.cs
   jng	    .loopss

   movss    xmm1,.t
   addss    xmm1,.dt
   movss    .t,xmm1
   pop	    ecx
   add	    ecx,1
   cmp	    ecx,.ct
   jng	    .looptt

   ; vertices described one bezier surface calculated


  ; calc triangles list

   mov	  edx,.ct
   mov	  esi,ebx
   xor	  ebx,ebx
 .oop1:
   mov	  ecx,.cs
  @@:
   mov	  ax,bx
   mov	  [esi],ax
   add	  ax,1
   mov	  [esi+10],ax
   mov	  [esi+4],ax
   add	  eax,.cs
   mov	  [esi+2],ax
   mov	  [esi+6],ax
   add	  ax,1
   mov	  [esi+8],ax
   add	  esi,12
   add	  bx,1
   add	  .tri_count,2
   sub	  ecx,1
   jnz	  @b
   add	  bx,1

   sub	  dx,1
   jnz	  .oop1



;   pop      ebx
;   pop      edi

   mov	     ebx,.tri_count
   imul      ebx,[i6]
   pop	     ecx
   add	     ebx,ecx
  ; ebx - end of computed triangles list ptr
  ; edi - end of calculated vertices ptr

   mov	    ecx,.vert_count
   mov	    edx,.tri_count

   add	    esp,512
   pop	    ebp

ret
if 0
draw_bezier_derives:	  ; bezier surface
  ; esi - source  derives
  ; edi - screen buffer
     movzx    eax,[xres_var]
     mov      [screen_width_var],eax
     mov      dword[lbuffer_ptr],edi
     mov      [line_color],0x0000ff00
     movd     xmm7,dword[xres_var]
     mov      eax,1
     movd     xmm5,eax
     shufps   xmm5,xmm5,0
     pxor     xmm6,xmm6
     punpcklwd xmm7,xmm6
     movlhps  xmm7,xmm7
     psubd    xmm7,xmm5
     push     esi
     mov      ecx,4
  .again1:
     push     ecx
     mov      ecx,3
  .again:
     push     ecx
     movlps   xmm0,[esi]
     movhps   xmm0,[esi+12]
     movaps   xmm3,xmm0
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip1

     ; eax=x1 ebx=y1 ecx=x2 edx=y2






     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     pushad
     call     linea
     popad
;     movhlps  xmm0,xmm3
;     movzx    edx,[yres_var]
;     movzx    edi,[yres_var]
;     sub      edx,4
;     sub      edi,4
;     push     esi
;     push     edx
;     push     edi
;     call     bar_edge
;     pop      edi
;     pop      edx

;     pop      esi
   .skip1:
     add     esi,12
     pop     ecx
     sub     ecx,1
     jnz     .again
     add     esi,12
     pop     ecx
     sub     ecx,1
     jnz     .again1
     pop     esi

     mov      ecx,4
  .again2:
     push     ecx
     mov      ecx,3
  .again3:
     push     ecx
     movlps   xmm0,[esi]
     movhps   xmm0,[esi+48]
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip2

     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     pushad
     call     linea
     popad
   .skip2:
     add     esi,12
     pop     ecx
     sub     ecx,1
     jnz     .again3
     pop     ecx
     sub     ecx,1
     jnz     .again2

ret
end if
draw_bezier_derives1:	   ; bezier surface
  ; esi - source  derives nodes
  ; ebx - souurce derv indexes (rectangles)
  ; edi - screen buffer
  ; uses some globals
     push     ebp
     mov      ebp,esp
     sub      esp,16
     .counter equ dword[ebp-4]
     mov      .counter,0
     movzx    eax,[xres_var]
     mov      [screen_width_var],eax
     mov      dword[lbuffer_ptr],edi
     mov      [line_color],0x0000ff00
     movd     xmm7,dword[xres_var]
     mov      eax,1
     movd     xmm5,eax
     shufps   xmm5,xmm5,0
     pxor     xmm6,xmm6
     punpcklwd xmm7,xmm6
     movlhps  xmm7,xmm7
     psubd    xmm7,xmm5
;     push     esi
     push     ebx
     mov      ecx,4
  .again1:
     push     ebx
     push     ecx
     mov      ecx,3
  .again:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]  ;.counter
     movzx    edx,word[ebx+2] ; .counter
     imul     edx,[i12]
     add      edx,esi
     imul     eax,[i12]
     add      eax,esi
     movlps   xmm0,[eax]
     movhps   xmm0,[edx]
     movaps   xmm3,xmm0
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip1


     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     pushad
     call     linea
     popad
;     movhlps  xmm0,xmm3
;     movzx    edx,[yres_var]
;     movzx    edi,[yres_var]
;     sub      edx,4
;     sub      edi,4
;     push     esi
;     push     edx
;     push     edi
;     call     bar_edge
;     pop      edi
;     pop      edx

;     pop      esi
   .skip1:
     pop     ebx
     add     ebx,2
     pop     ecx
     sub     ecx,1
     jnz     .again

     pop     ecx
     pop     ebx
     add     ebx,8
     sub     ecx,1
     jnz     .again1

     pop     ebx



     mov      ecx,4
  .again2:
     push     ecx
     mov      ecx,3
  .again3:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]
     movzx    edx,word[ebx+8]
     imul     eax,[i12]
     imul     edx,[i12]
     movlps   xmm0,[esi+eax]
     movhps   xmm0,[esi+edx]
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip2

     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     pushad
     call     linea
     popad
   .skip2:
     pop     ebx
     add     ebx,2
     pop     ecx
     sub     ecx,1
     jnz     .again3
     pop     ecx
     sub     ecx,1
     jnz     .again2

     mov   esp,ebp
     pop   ebp

ret

draw_bezier_bars:
; in:
;  esi - ptr to derives vertices list
;  ebx - ptr to derives indexes (16 pro one rectangle)
;  ecx - number of surfaces
;  globals - screen_ptr, i12, ..
   push  ebp
   mov	 ebp, esp
   .ym4  equ dword[ebp-4]
   sub	 esp,16

   shl	    ecx,4
   movzx    edx,[xres_var]
   movzx    edi,[yres_var]
   sub	    edx,4
   sub	    edi,4
   mov	    .ym4,edi
   pxor     xmm0,xmm0
 @@:
   push  ecx
   movzx eax,word[ebx]
   movzx ecx,word[ebx]
   imul  eax,[i12]
   add	 eax,esi
;   movss xmm0,[eax+8]
;   movlhps xmm0,xmm0
   movlps xmm0,[eax]
   push   ebx
   push   edx
   push   esi
   mov	  esi,.ym4
   call   bar_edge
   pop	  esi
   pop	  edx
   pop	  ebx
   add	  ebx,2
   pop	  ecx
   sub	  ecx,1
   jnz	  @b

   mov	  esp,ebp
   pop	  ebp
ret
if 0
do_derives_lists:
;  in: esi - ptr to derives
;      edi - derive vertices list
;      ebx - derive indexes list
;      ecx - number of surfaces
;out:  files structs in edi, ebx
   pushad
   push   ecx
   shl	  ecx,2
@@:
   movups xmm0,[esi]
   movups xmm1,[esi+16]
   movups xmm2,[esi+32]
   movups [edi],xmm0
   movups [edi+16],xmm1
   movups [edi+32],xmm2
   add	  esi,48
   add	  edi,48
   sub	  ecx,1
   jnz	  @b
   pop	  ecx
   shl	  ecx,4
   xor	  eax,eax
 @@:
   mov	  [ebx],ax
   add	  eax,1
   add	  ebx,2
   sub	  ecx,1
   jnz	  @b
   popad
ret
end if










