; Calculating surfaces based on b-splines
; by Maciej Guba, http://macgub.hekko.pl
; I use this equatins:
; x(s,t)=1/6*(1-3*t+3*t^2-t^3)*(a*x11+b*x12+c*x13+d*x14)+
;	+1/6*(4-6*t^2+3*t^3)*(a*x21+b*x22+c*x23+d*x24)+
;	+1/6*(1+3*t+3*t^2-3*t^3)*(a*x31+b*x32+c*x33+d*x34)+
;	+1/6*t^3*(a*x41+b*x42+c*x43+d*x44)
; a =	1/6*(1-3*s+3*s^2-s^3)
; b =	1/6*(4-6*s^2+3*s^3)
; c =	1/6*(1+3*s+3*s^2-3*s^3)
; d =	1/6*s^3
; att =   1/6*(1-3*t+3*t^2-t^3)
; btt =   1/6*(4-6*t^2+3*t^3)
; ctt =   1/6*(1+3*t+3*t^2-3*t^3)
; dtt =   1/6*t^3
calc_bssurface:
; calc single surface, write vertices to [edi],
; faces to [ebx]
;  in  :
;	esi - ptr  -> x11, y11, z11 .... x44, y44, z44 ; dword
;	edx - quality -> s shl 16 t
;	edi - ptr to area of calculated vertices
;	ebx - ptr to area of calculated faces (triangles)
;  out :
;	ecx - vertices count
;	edx - triangles count

	push  ebp
	mov   ebp,esp
	sub   esp,512
	sub   ebp,16
	and   ebp,0xfffffff0

  .t equ [ebp-4]
  .s equ [ebp-8]
  .ct equ	  dword [ebp-12] ; integer counter t
  .cs equ	  dword [ebp-16] ; int counter s
  .vert_count equ dword [ebp-20]
  .tri_count  equ dword [ebp-24]

  .x4 equ ebp-48	;      .x11-x14
  .x3 equ ebp-64	;      .x21-x24
  .x2 equ ebp-80	;      .x31-x34
  .x1 equ ebp-96	;      .x41-x44
  .y4 equ ebp-112	;      .y11-y14
  .y3 equ ebp-128	;      .y21-y24
  .y2 equ ebp-144	;      .y31-y34
  .y1 equ ebp-160	;      .y41-y44
  .z4 equ ebp-176	;      .z11-z14
  .z3 equ ebp-192	;      .z21-z24
  .z2 equ ebp-208	;      .z31-z34
  .z1 equ ebp-224	;      .z41-z44
  .ds equ  [ebp-228]	; delta s
  .dt equ  [ebp-232]	; delta t
  .d  equ  [ebp-244]
  .c  equ  [ebp-248]
  .b  equ  [ebp-252]
  .a  equ  [ebp-256]
  .dtt equ [ebp-260]
  .ctt equ [ebp-264]
  .btt equ [ebp-268]
  .att equ [ebp-272]


   push dword[esi]
   pop	dword[.x1]
   push dword[esi+12]
   pop	dword[.x1+4]
   push dword[esi+24]
   pop	dword[.x1+8]
   push dword[esi+36]
   pop	dword[.x1+12]

   push dword[esi+36+12]    ; shuffle x
   pop	dword[.x2]
   push dword[esi+36+12*2]
   pop	dword[.x2+4]
   push dword[esi+36+12*3]
   pop	dword[.x2+8]
   push dword[esi+36+12*4]
   pop	dword[.x2+12]

   push dword[esi+36+12*5]
   pop	dword[.x3]
   push dword[esi+36+12*6]
   pop	dword[.x3+4]
   push dword[esi+36+12*7]
   pop	dword[.x3+8]
   push dword[esi+36+12*8]
   pop	dword[.x3+12]

   push dword[esi+36+12*9]
   pop	dword[.x4]
   push dword[esi+36+12*10]
   pop	dword[.x4+4]
   push dword[esi+36+12*11]
   pop	dword[.x4+8]
   push dword[esi+36+12*12]
   pop	dword[.x4+12]

   push dword[esi+4]	      ;shuffle y
   pop	dword[.y1]
   push dword[esi+12+4]
   pop	dword[.y1+4]
   push dword[esi+24+4]
   pop	dword[.y1+8]
   push dword[esi+36+4]
   pop	dword[.y1+12]

   push dword[esi+36+12+4]
   pop	dword[.y2]
   push dword[esi+36+12*2+4]
   pop	dword[.y2+4]
   push dword[esi+36+12*3+4]
   pop	dword[.y2+8]
   push dword[esi+36+12*4+4]
   pop	dword[.y2+12]

   push dword[esi+36+12*5+4]
   pop	dword[.y3]
   push dword[esi+36+12*6+4]
   pop	dword[.y3+4]
   push dword[esi+36+12*7+4]
   pop	dword[.y3+8]
   push dword[esi+36+12*8+4]
   pop	dword[.y3+12]

   push dword[esi+36+12*9+4]
   pop	dword[.y4]
   push dword[esi+36+12*10+4]
   pop	dword[.y4+4]
   push dword[esi+36+12*11+4]
   pop	dword[.y4+8]
   push dword[esi+36+12*12+4]
   pop	dword[.y4+12]


   push dword[esi+8]	      ;shuffle z
   pop	dword[.z1]
   push dword[esi+12+8]
   pop	dword[.z1+4]
   push dword[esi+24+8]
   pop	dword[.z1+8]
   push dword[esi+36+8]
   pop	dword[.z1+12]

   push dword[esi+36+12+8]
   pop	dword[.z2]
   push dword[esi+36+12*2+8]
   pop	dword[.z2+4]
   push dword[esi+36+12*3+8]
   pop	dword[.z2+8]
   push dword[esi+36+12*4+8]
   pop	dword[.z2+12]

   push dword[esi+36+12*5+8]
   pop	dword[.z3]
   push dword[esi+36+12*6+8]
   pop	dword[.z3+4]
   push dword[esi+36+12*7+8]
   pop	dword[.z3+8]
   push dword[esi+36+12*8+8]
   pop	dword[.z3+12]

   push dword[esi+36+12*9+8]
   pop	dword[.z4]
   push dword[esi+36+12*10+8]
   pop	dword[.z4+4]
   push dword[esi+36+12*11+8]
   pop	dword[.z4+8]
   push dword[esi+36+12*12+8]
   pop	dword[.z4+12]


   movzx    ecx,dx
   mov	    .cs,ecx
   cvtsi2ss xmm1,ecx
   rcpss    xmm1,xmm1
   movss    .ds,xmm1
   shr	    edx,16
   mov	    .ct,edx
   cvtsi2ss xmm0,edx
   rcpss    xmm0,xmm0
   movss    .dt,xmm0

   push     edi

   xor	    ecx,ecx
   mov	    .s,ecx
 ;  mov      .t,ecx
   mov	    .vert_count,ecx
   mov	    .tri_count,ecx
;   xorps    xmm2,xmm2
;   movq     .s,xmm2 ; zero both s and t

.oops:
   push     ecx
   movss    xmm1,.s
   movss    xmm2,xmm1
   mulss    xmm2,xmm1
   movss    xmm3,xmm2
   mulss    xmm3,xmm1
   movss    xmm0,xmm3
   movlhps  xmm0,xmm2
   shufps   xmm0,xmm0,11111000b
   movlhps  xmm0,xmm1
   addps    xmm0,dqword[mask1000]

   movaps   xmm4,xmm0
   movaps   xmm5,xmm0
   movaps   xmm6,xmm0
   movss    xmm7,xmm0
   mulps    xmm4,dqword[mul1m33m1]
   mulps    xmm5,dqword[mul40m63]
   mulps    xmm6,dqword[mul133m3]
   mulss    xmm7,[recipr6]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   mulss    xmm4,[recipr6] ; lowest dword xmm4 - a
   mulss    xmm5,[recipr6]   ;lo xmm5 -  b
   mulss    xmm6,[recipr6]   ;lo xmm5 -  b
   movss    .a,xmm4
   movss    .b,xmm5
   movss    .c,xmm6
   movss    .d,xmm7

   xor	    ecx,ecx
   mov	    dword .t,ecx
 .oopt:
   push     ecx

   movss    xmm1,.t
   movss    xmm2,xmm1
   mulss    xmm2,xmm1
   movss    xmm3,xmm2
   mulss    xmm3,xmm1
   movss    xmm0,xmm3
   movlhps  xmm0,xmm2
   shufps   xmm0,xmm0,11111000b
   movlhps  xmm0,xmm1

   movaps   xmm4,xmm0
   movaps   xmm5,xmm0
   movaps   xmm6,xmm0
   movss    xmm7,xmm0
   mulps    xmm4,dqword[mul1m33m1]
   mulps    xmm5,dqword[mul40m63]
   mulps    xmm6,dqword[mul133m3]
   mulss    xmm7,[recipr6]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   mulss    xmm4,[recipr6] ; lowest dword xmm4 - a
   mulss    xmm5,[recipr6]   ;lo xmm5 -  b
   mulss    xmm6,[recipr6]   ;lo xmm5 -  b
 ;  movss    .att,xmm4
 ;  movss    .btt,xmm5
 ;  movss    .ctt,xmm6
 ;  movss    .dtt,xmm7


   movaps   xmm0,.a
   movaps   xmm1,xmm0
   movaps   xmm2,xmm0
   movaps   xmm3,xmm0
   mulps    xmm1,[.x1]
   mulps    xmm2,[.x2]
   mulps    xmm3,[.x3]
   mulps    xmm0,[.x4]
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   mulss    xmm1,xmm4	 ; att
   mulss    xmm2,xmm5	 ; btt
   mulss    xmm3,xmm6	 ; ctt
   mulss    xmm0,xmm7	 ; dtt
   addss    xmm1,xmm2
   addss    xmm3,xmm0
   addss    xmm1,xmm3 ; lowest dword of xmm1 - desired x
   movss    [edi],xmm1


   movaps   xmm0,.a
   movaps   xmm1,xmm0
   movaps   xmm2,xmm0
   movaps   xmm3,xmm0
   mulps    xmm1,[.y1]
   mulps    xmm2,[.y2]
   mulps    xmm3,[.y3]
   mulps    xmm0,[.y4]
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   mulss    xmm1,xmm4  ;.att  ;xmm4
   mulss    xmm2,xmm5  ;.btt  ;xmm5
   mulss    xmm3,xmm6  ;.ctt  ;xmm6
   mulss    xmm0,xmm7  ;.dtt  ;xmm7
   addss    xmm1,xmm2
   addss    xmm3,xmm0
   addss    xmm1,xmm3 ; lowest dword of xmm1 - desired y
   movss    [edi+4],xmm1



   movaps   xmm0,.a
   movaps   xmm1,xmm0
   movaps   xmm2,xmm0
   movaps   xmm3,xmm0
   mulps    xmm1,[.z1]
   mulps    xmm2,[.z2]
   mulps    xmm3,[.z3]
   mulps    xmm0,[.z4]
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   haddps   xmm1,xmm1
   haddps   xmm2,xmm2
   haddps   xmm3,xmm3
   haddps   xmm0,xmm0
   mulss    xmm1,xmm4 ;.att  ;xmm4
   mulss    xmm2,xmm5 ;.btt  ;xmm5
   mulss    xmm3,xmm6 ;.ctt  ;xmm6
   mulss    xmm0,xmm7 ;.dtt  ;xmm7
   addss    xmm1,xmm2
   addss    xmm3,xmm0
   addss    xmm1,xmm3 ; lowest dword of xmm1 - desired z
   movss    [edi+8],xmm1


   movss    xmm0,.t
   addss    xmm0,.dt
   movss    .t,xmm0
   add	    edi,12
   add	    .vert_count,1
   pop	    ecx
   add	    ecx,1
   cmp	    ecx,.ct
   jng	    .oopt

   movss    xmm1,.s
   addss    xmm1,.ds
   movss    .s,xmm1
   pop	    ecx
   add	    ecx,1
   cmp	    ecx,.cs
   jng	    .oops

   ; vertices described one surface calculated

  ; pop      edi
   push     ebx
if 1

  ; calc triangles list
   xor	    eax,eax
   mov	    ecx,.cs
   add	    ecx,1
   mov	    edx,.ct
   imul     edx,ecx
   sub	    edx,3
   xor	    edi,edi
;   sub      .cs,1
;   sub      .ct,1
   mov	    edi,-1
.tri_loop1:
   add	    edi,.cs
  ; sub      edi,1
.tri_loop:
   mov	    [ebx],ax
   add	    ax,1
   mov	    [ebx+2],ax
   mov	    [ebx+4],cx

   mov	    [ebx+6],ax
   mov	    [ebx+8],cx
   add	    cx,1
   mov	    [ebx+10],cx

   add	    ebx,12
   add	    .tri_count,2
   cmp	    eax,edi
   jna	   .tri_loop
  ; add      ax,2
  ; add      cx,2
   cmp	    cx,dx
   jnae     .tri_loop1

end if

if 0

   mov	  edx,.cs
   sub	  edx,1
   mov	  esi,ebx
   xor	  ebx,ebx
 .oop1:
   mov	  ecx,.ct
   sub	  ecx,1
  @@:
   mov	  ax,bx
   mov	  [esi],ax
   add	  ax,1
   mov	  [esi+10],ax
   mov	  [esi+2],ax
   add	  eax,.ct
   sub	  ax,1
   mov	  [esi+4],ax
   mov	  [esi+6],ax
   add	  ax,1
   mov	  [esi+8],ax
   add	  esi,12
   inc	  bx
   add	  .tri_count,2
   sub	  ecx,1
   jnz	  @b
   add	  bx,1

   sub	  dx,1
   cmp	  dx,0
   jnz	  .oop1

end if










   pop	    ebx
   pop	    edi

   mov	    ecx,.vert_count
   mov	    edx,.tri_count

   add	    esp,512
   pop	    ebp

ret


